# -*- coding: utf-8 -*-
#
# Copyright 2020 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import datetime
import logging
from typing import Any, Dict, Iterable, List, Type

from .base import KelvinObject, KelvinResource
from .exceptions import InvalidRequest
from .session import Session

logger = logging.getLogger(__name__)


class User(KelvinObject):
    _class_display_name = "User"
    _kelvin_attrs = KelvinObject._kelvin_attrs + [
        "school",
        "firstname",
        "lastname",
        "birthday",
        "disabled",
        "email",
        "password",
        "record_uid",
        "roles",
        "schools",
        "school_classes",
        "source_uid",
        "udm_properties",
    ]

    def __init__(
        self,
        name: str,
        school: str,
        *,
        firstname: str = None,
        lastname: str = None,
        birthday: datetime.date = None,
        disabled: bool = False,
        email: str = None,
        password: str = None,
        record_uid: str = None,
        roles: List[str],
        schools: List[str],
        school_classes: Dict[str, List[str]] = None,
        source_uid: str = None,
        udm_properties: Dict[str, Any] = None,
        ucsschool_roles: List[str] = None,
        dn: str = None,
        url: str = None,
        session: Session = None,
    ):
        super().__init__(
            name=name, ucsschool_roles=ucsschool_roles, dn=dn, url=url, session=session
        )
        self.school = school
        self.firstname = firstname
        self.lastname = lastname
        self.birthday = birthday
        self.disabled = disabled
        self.email = email
        self.password = password
        self.record_uid = record_uid
        self.roles = roles
        self.schools = schools
        self.school_classes = school_classes or {}
        self.source_uid = source_uid
        self.udm_properties = udm_properties or {}
        self._resource_class = UserResource

    @classmethod
    def _from_kelvin_response(cls, response: Dict[str, Any]) -> "User":
        for attr in ("roles", "schools"):
            # turn urls to names
            response[attr] = [url.rsplit("/", 1)[-1] for url in response[attr]]
        # 'school' will be done in super class
        return super()._from_kelvin_response(response)

    def _to_kelvin_request_data(self) -> Dict[str, Any]:
        data = super()._to_kelvin_request_data()
        url_keys = {"roles": "role", "schools": "school"}
        for attr in ("roles", "schools"):
            url_key = url_keys[attr]
            # role/school names to urls
            data[attr] = [
                f"{self.session.urls[url_key]}{value}" for value in data[attr]
            ]
        return data


class UserResource(KelvinResource):
    class Meta:
        kelvin_object: Type[KelvinObject] = User
        required_get_attrs: Iterable[str] = ("name",)
        required_search_attrs: Iterable[str] = ()

    def __init__(self, session: Session):
        super().__init__(session=session)
        self.collection_url = self.session.urls["user"]
        self.object_url = f"{self.session.urls['user']}{{name}}"

    def _check_search_attrs(self, **kwargs) -> None:
        super()._check_search_attrs(**kwargs)
        if "*" in kwargs.get("school", ""):
            raise InvalidRequest(
                f"Argument 'school' for searching users must be exact."
            )
